from __future__ import annotations

import json
from typing import Any, Dict, Union, Optional, Sequence, cast
from typing_extensions import override

import tiktoken  # type: ignore

from payi.lib.helpers import PayiCategories
from payi.types.ingest_units_params import IngestUnits

from .instrument import (
    PayiInstrumentOpenAiAzureConfig,
    _Context,
    _IsStreaming,
    _model_to_dict,
    _PayiInstrumentor,
)
from .version_helper import get_version_helper
from .ProviderRequest import _ChunkResult, _StreamingType, _ProviderRequest


class OpenAiInstrumentor:
    _module_name: str = "openai"
    _module_version: str = ""

    _azure_openai_deployments: Dict[str, _Context] = {}

    @staticmethod
    def is_azure(openai_client: Any) -> bool:
        from openai import AzureOpenAI, AsyncAzureOpenAI # type: ignore # noqa: I001

        return isinstance(openai_client, (AsyncAzureOpenAI, AzureOpenAI))

    @staticmethod
    def configure(azure_openai_config: Optional[PayiInstrumentOpenAiAzureConfig]) -> None:
        if azure_openai_config:
            model_mappings = azure_openai_config.get("model_mappings", [])
            OpenAiInstrumentor._azure_openai_deployments = _PayiInstrumentor._model_mapping_to_context_dict(model_mappings)

    @staticmethod
    def assign_module_version() -> None:
        if not OpenAiInstrumentor._module_version:
            OpenAiInstrumentor._module_version = get_version_helper(OpenAiInstrumentor._module_name)
    @staticmethod
    def instrument(instrumentor: _PayiInstrumentor) -> None:
        OpenAiInstrumentor.assign_module_version()

        wrappers = [
            ("openai._base_client", "AsyncAPIClient._process_response", _ProviderRequest.aprocess_response_wrapper),
            ("openai._base_client", "SyncAPIClient._process_response", _ProviderRequest.process_response_wrapper),
            ("openai.resources.chat.completions", "Completions.create", chat_wrapper(instrumentor)),
            ("openai.resources.chat.completions", "Completions.parse", chat_wrapper(instrumentor)),
            ("openai.resources.chat.completions", "AsyncCompletions.create", achat_wrapper(instrumentor)),
            ("openai.resources.chat.completions", "AsyncCompletions.parse", achat_wrapper(instrumentor)),
            ("openai.resources.embeddings", "Embeddings.create", embeddings_wrapper(instrumentor)),
            ("openai.resources.embeddings", "AsyncEmbeddings.create", aembeddings_wrapper(instrumentor)),
            ("openai.resources.responses", "Responses.create", responses_wrapper(instrumentor)),
            ("openai.resources.responses", "Responses.parse", responses_wrapper(instrumentor)),
            ("openai.resources.responses", "AsyncResponses.create", aresponses_wrapper(instrumentor)),
            ("openai.resources.responses", "AsyncResponses.parse", aresponses_wrapper(instrumentor)),

            # In post beta openai moddule releases wrapping these will fail and gracefully handled
            ("openai.resources.beta.chat.completions", "Completions.create", chat_wrapper(instrumentor)),
            ("openai.resources.beta.chat.completions", "Completions.parse", chat_wrapper(instrumentor)),
            ("openai.resources.beta.chat.completions", "AsyncCompletions.create", achat_wrapper(instrumentor)),
            ("openai.resources.beta.chat.completions", "AsyncCompletions.parse", achat_wrapper(instrumentor)),
        ]

        instrumentor._wrap_functions(wrappers)

    @staticmethod
    def instrument_databricks(instrumentor: _PayiInstrumentor) -> None:
        OpenAiInstrumentor.assign_module_version()
    
        wrappers = [
            ("databricks.sdk.service.serving", "ServingEndpointsAPI.query", databricks_query_wrapper(instrumentor)),
            ("databricks.sdk.service.serving", "ServingEndpointsDataPlaneAPI.query", databricks_query_wrapper(instrumentor)),
        ]

        instrumentor._wrap_functions(wrappers)

@_PayiInstrumentor.payi_wrapper
def embeddings_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("OpenAI Embeddings wrapper")
    return instrumentor.invoke_wrapper(
        _OpenAiEmbeddingsProviderRequest(instrumentor, instance),
        _IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
async def aembeddings_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("async OpenAI Embeddings wrapper")
    return await instrumentor.async_invoke_wrapper(
        _OpenAiEmbeddingsProviderRequest(instrumentor, instance),
        _IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
def chat_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("OpenAI completions wrapper")
    return instrumentor.invoke_wrapper(
        _OpenAiChatProviderRequest(instrumentor, instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_awrapper
async def achat_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("async OpenAI completions wrapper")
    return await instrumentor.async_invoke_wrapper(
        _OpenAiChatProviderRequest(instrumentor, instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )
    
@_PayiInstrumentor.payi_wrapper
def responses_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("OpenAI responses wrapper")
    return instrumentor.invoke_wrapper(
        _OpenAiResponsesProviderRequest(instrumentor, instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_awrapper
async def aresponses_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("async OpenAI responses wrapper")
    return await instrumentor.async_invoke_wrapper(
        _OpenAiResponsesProviderRequest(instrumentor, instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
def databricks_query_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("Databricks query wrapper")
    return instrumentor.invoke_wrapper(
        _DatabricksOpenAiChatProviderRequest(instrumentor, instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

class _OpenAiProviderRequest(_ProviderRequest):
    chat_input_tokens_key: str = "prompt_tokens"
    chat_output_tokens_key: str = "completion_tokens"
    chat_input_tokens_details_key: str = "prompt_tokens_details"
    chat_completion_tokens_details_key: str = "completion_tokens_details"

    responses_input_tokens_key: str = "input_tokens"
    responses_output_tokens_key: str = "output_tokens"
    responses_input_tokens_details_key: str = "input_tokens_details"
    responses_output_tokens_details_key: str = "output_tokens_details"

    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any, input_tokens_key: str, output_tokens_key: str, input_tokens_details_key: str, output_tokens_details_key: str) -> None:
        self._openai_client = instance._client if instance and hasattr(instance, "_client") else None
        self._is_azure = self._openai_client and OpenAiInstrumentor.is_azure(self._openai_client)

        category = PayiCategories.azure_openai if self._is_azure else PayiCategories.openai

        super().__init__(
            instrumentor=instrumentor,
            category=category,
            streaming_type=_StreamingType.iterator,
            module_name=OpenAiInstrumentor._module_name,
            module_version=OpenAiInstrumentor._module_version,
            )
        self._input_tokens_key = input_tokens_key
        self._output_tokens_key = output_tokens_key
        self._input_tokens_details_key = input_tokens_details_key
        self._output_tokens_details_key = output_tokens_details_key

        if hasattr(self._openai_client, "base_url"):
           try:
               self.provider_uri = str(self._openai_client.base_url) # type: ignore
           except Exception:
               pass

    def get_model(self, kwargs: Any) -> str:
        return kwargs.get("model", "") or ""

    def get_messages(self, kwargs: Any) -> Optional[list[Dict[str, Any]]]:
        return cast(Optional[list[Dict[str, Any]]], kwargs.get("messages", None))

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool: # type: ignore
        model = self.get_model(kwargs)

        if self._is_azure:
            # model is technically optional as it is part of the URL path
            if not model and hasattr(self._openai_client, "_azure_deployment"):
                model = self._openai_client._azure_deployment # type: ignore

            self.map_deployment(model)

        self.apply_price_as(model)

        return True

    def apply_price_as(self, model: str) -> None:
        if self._price_as.resource_scope:
            self._ingest["resource_scope"] = self._price_as.resource_scope

        if self._price_as.category:
            self._category = self._price_as.category

        self._ingest["category"] = self._category
        self._ingest["resource"] = self._price_as.resource if self._price_as.resource else model

    def map_deployment(self, model:Optional[str]) -> None:
        self._instrumentor._logger.debug(f"Azure OpenAI model {model}, available mappings: {list(OpenAiInstrumentor._azure_openai_deployments.keys())}, price as before final mapping: resource={self._price_as.resource}, category={self._price_as.category}, resource_scope={self._price_as.resource_scope}")

        if model and not self._price_as.resource and not self._price_as.category and model in OpenAiInstrumentor._azure_openai_deployments:
            deployment = OpenAiInstrumentor._azure_openai_deployments.get(model, {})
            self._price_as.category = deployment.get("price_as_category", None)
            self._price_as.resource = deployment.get("price_as_resource", None)
            self._price_as.resource_scope = deployment.get("resource_scope", None)

        if not self._price_as.resource and not self._price_as.category:
            self._instrumentor._logger.warning("Azure OpenAI requires price as resource and/or category to be specified unless mapped in the Pay-i service")

    @override
    def process_exception(self, exception: Exception, kwargs: Any, ) -> bool:
        try:
            status_code: Optional[int] = None

            if hasattr(exception, "status_code"):
                status_code = getattr(exception, "status_code", None)
                if isinstance(status_code, int):
                    self._ingest["http_status_code"] = status_code

            if not status_code:
                self.exception_to_semantic_failure(exception,)
                return True

            if hasattr(exception, "request_id"):
                request_id = getattr(exception, "request_id", None)
                if isinstance(request_id, str):
                    self._ingest["provider_response_id"] = request_id

            if hasattr(exception, "response"):
                response = getattr(exception, "response", None)
                if hasattr(response, "text"):
                    text = getattr(response, "text", None)
                    if isinstance(text, str):
                        self._ingest["provider_response_json"] = text

        except Exception as e:
            self._instrumentor._logger.debug(f"Error processing exception: {e}")
            return False

        return True

    def update_model_name(self, model_name: str) -> None:
        resource = self._ingest.get("resource", None)

        if (resource is None or len(resource) == 0) and len(model_name) > 0:
            self.map_deployment(model_name)
            self.apply_price_as(model_name)

    def update_deployment_name(self) -> bool:
        if self._ingest.get("resource", None):
            return False

        deployment_name = self.find_response_header_value("x-ms-deployment-name")
        if deployment_name:
            self.map_deployment(deployment_name)
            self.apply_price_as(deployment_name)
            return True

        return False

    def process_synchronous_response_worker(
        self,
        response_dict: dict[str, Any],
        ) -> Any:

        if self.update_deployment_name() is False:
            self.update_model_name(response_dict.get("model", ""))
        
        self.add_usage_units(response_dict.get("usage", {}))

        if self._log_prompt_and_response:
            self._ingest["provider_response_json"] = [json.dumps(response_dict)]

        if "id" in response_dict:
            self._ingest["provider_response_id"] = response_dict["id"]

        return None

    def add_usage_units(self, usage: "dict[str, Any]",) -> None:
        units = self._ingest["units"]

        input = usage[self._input_tokens_key] if self._input_tokens_key in usage else 0
        output = usage[self._output_tokens_key] if self._output_tokens_key in usage else 0
        input_cache = 0

        prompt_tokens_details = usage.get(self._input_tokens_details_key, {})
        if prompt_tokens_details:
            input_cache = prompt_tokens_details.get("cached_tokens", 0)
            if input_cache != 0:
                units["text_cache_read"] = IngestUnits(input=input_cache, output=0)

        output_tokens_details = usage.get(self._output_tokens_details_key, {})
        if output_tokens_details:
            reasoning_tokens = output_tokens_details.get("reasoning_tokens", 0)
            if reasoning_tokens != 0:
                units["reasoning"] = IngestUnits(input=0, output=reasoning_tokens)
                output -= reasoning_tokens

        input = self.update_for_vision(input - input_cache)

        units["text"] = IngestUnits(input=input, output=output)

    @staticmethod
    def has_image_and_get_texts(encoding: tiktoken.Encoding, content: Union[str, 'list[Any]'], image_type: str, text_type: str) -> 'tuple[bool, int]':
        if isinstance(content, list): # type: ignore
            has_image = any(item.get("type", "") == image_type for item in content)
            if has_image is False:
                return has_image, 0
            
            token_count = sum(len(encoding.encode(item.get("text", ""))) for item in content if item.get("type") == text_type)
            return has_image, token_count
        return False, 0

    @staticmethod
    def post_process_request_prompt(content: Union[str, 'list[Any]'], image_type: str, url_subkey: bool) -> bool:
        modified = False
        if isinstance(content, list): # type: ignore
            for item in content:
                type = item.get("type", "")
                if type != image_type:
                    continue

                if url_subkey:
                    url = item.get("image_url", {}).get("url", "")
                    if url.startswith("data:"):
                        item["image_url"]["url"] = _PayiInstrumentor._not_instrumented
                        modified = True
                else:
                    url = item.get("image_url", "")
                    if url.startswith("data:"):
                        item["image_url"] = _PayiInstrumentor._not_instrumented
                        modified = True

        return modified

class _OpenAiEmbeddingsProviderRequest(_OpenAiProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any):
        super().__init__(
            instrumentor=instrumentor,
            instance=instance,
            input_tokens_key=_OpenAiProviderRequest.chat_input_tokens_key,
            output_tokens_key=_OpenAiProviderRequest.chat_output_tokens_key,
            input_tokens_details_key=_OpenAiProviderRequest.chat_input_tokens_details_key,
            output_tokens_details_key=_OpenAiProviderRequest.chat_completion_tokens_details_key)

    @override
    def process_synchronous_response(
        self,
        response: Any,
        kwargs: Any) -> Any:
        return self.process_synchronous_response_worker(_model_to_dict(response))

class _OpenAiChatProviderRequest(_OpenAiProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any):
        super().__init__(
            instrumentor=instrumentor,
            instance=instance,
            input_tokens_key=_OpenAiProviderRequest.chat_input_tokens_key,
            output_tokens_key=_OpenAiProviderRequest.chat_output_tokens_key,
            input_tokens_details_key=_OpenAiProviderRequest.chat_input_tokens_details_key,
            output_tokens_details_key=_OpenAiProviderRequest.chat_completion_tokens_details_key)

        self._include_usage_added = False

    @staticmethod
    def process_synchronous_response_choices(request: _ProviderRequest, response: dict[str, Any]) -> None:
        choices = response.get("choices", [])
        if choices:
            for choice in choices:
                function = choice.get("message", {}).get("function_call", {})

                if not function:
                    continue

                name = function.get("name", None)
                arguments = function.get("arguments", None)

                if name:
                    request.add_synchronous_function_call(name=name, arguments=arguments)

    @staticmethod
    def process_chunk_choices(request: _ProviderRequest, chunk: dict[str, Any]) -> None:
        choices = chunk.get("choices", [])
        if choices:
            for choice in choices:
                function = choice.get("delta", {}).get("function_call", {})
                index = choice.get("index", None)

                if function and index is not None:
                    name = function.get("name", None)
                    arguments = function.get("arguments", None)

                    if name or arguments:
                        request.add_streaming_function_call(index=index, name=name, arguments=arguments)

    def add_stream_options(self) -> bool:
        return True
    
    @override
    def process_chunk(self, chunk: Any) -> _ChunkResult:
        ingest = False
        model = _model_to_dict(chunk)
        
        if "provider_response_id" not in self._ingest:
            response_id = model.get("id", None)
            if response_id:
                self._ingest["provider_response_id"] = response_id

        send_chunk_to_client = True

        self.process_chunk_choices(self, model)

        usage = model.get("usage", {})
        if usage:
            if self.update_deployment_name() is False:
                self.update_model_name(model.get("model", ""))                
            self.add_usage_units(usage)

            # If we added "include_usage" in the request on behalf of the client, do not return the extra 
            # packet which contains the usage to the client as they are not expecting the data
            if self._include_usage_added:
                send_chunk_to_client = False
            ingest = True

        return _ChunkResult(send_chunk_to_caller=send_chunk_to_client, ingest=ingest)

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        result = super().process_request(instance, args, kwargs)
        if result is False:
            return result
        
        messages = self.get_messages(kwargs)
        if messages:
            estimated_token_count = 0 
            has_image = False
            enc: Optional[tiktoken.Encoding] = None

            try: 
                enc = tiktoken.encoding_for_model(self.get_model(kwargs)) # type: ignore
            except Exception:
                try:
                    enc = tiktoken.get_encoding("o200k_base") # type: ignore
                except Exception:
                    self._instrumentor._logger.info("OpenAI skipping vision token calc, could not load o200k_base")
                    enc = None
            
            if enc:
                for message in messages:
                    msg_has_image, msg_prompt_tokens = self.has_image_and_get_texts(enc, message.get('content', ''), image_type="image_url", text_type="text")
                    if msg_has_image:
                        has_image = True
                        estimated_token_count += msg_prompt_tokens
            
                if has_image and estimated_token_count > 0:
                    self._estimated_prompt_tokens = estimated_token_count

            stream: bool = kwargs.get("stream", False)
            if stream and self.add_stream_options():
                add_include_usage = True

                stream_options: dict[str, Any] = kwargs.get("stream_options", None)
                if stream_options and "include_usage" in stream_options:
                    add_include_usage = stream_options["include_usage"] == False

                if add_include_usage:
                    kwargs['stream_options'] = {"include_usage": True}
                    self._include_usage_added = True
        return True

    @override
    def remove_prompt_inline_data(self, prompt: 'dict[str, Any]') -> bool:
        messages = self.get_messages(prompt)
        if not messages:
            return False
        return self.post_process_request_prompt(messages, image_type="image_url", url_subkey=True)

    @override
    def process_synchronous_response(
        self,
        response: Any,
        kwargs: Any) -> Any:

        response_dict = _model_to_dict(response)
        self.process_synchronous_response_choices(self, response_dict)

        return self.process_synchronous_response_worker(response_dict)

class _DatabricksOpenAiChatProviderRequest(_OpenAiChatProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any):
        super().__init__(instrumentor=instrumentor, instance=instance)
    
        provider_uri = None
        try:
            provider_uri = str(instance._api._cfg.host)  # type: ignore
        except Exception:
            pass

        if not provider_uri:
            try:
                provider_uri = str(instance._api._cfg.hostname)  # type: ignore
            except Exception:
                pass
        
        if provider_uri:
            if 'https://' not in provider_uri:
                provider_uri = f"https://{provider_uri}"
            self.provider_uri = provider_uri

    @override
    def add_stream_options(self) -> bool:
        return False

    @override
    def get_model(self, kwargs: Any) -> str:
        return kwargs.get("name", "") or ""
    
    @override
    def get_messages(self, kwargs: Any) -> Optional[list[Dict[str, Any]]]:
        messages = kwargs.get("messages", None)
        if messages:
            result: list[Dict[str, Any]] = []
            for message in messages:
                result.append(cast(Dict[str, Any], message if isinstance(message, dict) else message.as_dict()))
            return result
        
        return None


class _OpenAiResponsesProviderRequest(_OpenAiProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any):
        super().__init__(
            instrumentor=instrumentor,
            instance=instance,
            input_tokens_key=_OpenAiProviderRequest.responses_input_tokens_key,
            output_tokens_key=_OpenAiProviderRequest.responses_output_tokens_key,
            input_tokens_details_key=_OpenAiProviderRequest.responses_input_tokens_details_key,
            output_tokens_details_key=_OpenAiProviderRequest.responses_output_tokens_details_key)

    @override
    def process_chunk(self, chunk: Any) -> _ChunkResult:
        ingest = False
        model = _model_to_dict(chunk)
        response: dict[str, Any] = model.get("response", {})

        if "provider_response_id" not in self._ingest:
            response_id = response.get("id", None)
            if response_id:
                self._ingest["provider_response_id"] = response_id

        type = model.get("type", "")
        if type and type == "response.output_item.done":
            item = model.get("item", {})
            if item and item.get("type", "") == "function_call":
                name = item.get("name", None)
                arguments = item.get("arguments", None)

                if name:
                    self.add_synchronous_function_call(name=name, arguments=arguments)

        usage = response.get("usage")
        if usage:
            if self.update_deployment_name() is False:
                self.update_model_name(response.get("model", ""))                

            self.add_usage_units(usage)
            ingest = True

        return _ChunkResult(send_chunk_to_caller=True, ingest=ingest)

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        result = super().process_request(instance, args, kwargs)
        if result is False:
            return result
        
        input = kwargs.get("input", None) # type: ignore
        if not input or isinstance(input, str) or not isinstance(input, list):
            return True
        
        estimated_token_count = 0 
        has_image = False
        enc: Optional[tiktoken.Encoding] = None

        try: 
            enc = tiktoken.encoding_for_model(kwargs.get("model")) # type: ignore
        except Exception:
            try:
                enc = tiktoken.get_encoding("o200k_base") # type: ignore
            except Exception:
                self._instrumentor._logger.info("OpenAI skipping vision token calc, could not load o200k_base")
                enc = None
        
        # find each content..type="input_text" and count tokens
        # input=[{
        #     "role": "user",
        #     "content": [
        #         {
        #             "type": "input_text",
        #             "text": "what's in this image?"
        #         },
        #         {
        #             "type": "input_image",
        #             "image_url": ... 
        #         },
        #     ],
        # }]
        if enc:
            for item in input: # type: ignore
                if isinstance(item, dict):
                    for key, value in item.items(): # type: ignore
                        if key == "content":
                            if isinstance(value, list):
                                msg_has_image, msg_prompt_tokens = self.has_image_and_get_texts(enc, value, image_type="input_image", text_type="input_text") # type: ignore 
                                if msg_has_image:
                                    has_image = True
                                    estimated_token_count += msg_prompt_tokens

            if has_image and estimated_token_count > 0:
                self._estimated_prompt_tokens = estimated_token_count

        return True

    @override
    def remove_prompt_inline_data(self, prompt: 'dict[str, Any]') -> bool:
        modified = False
        input = prompt.get("input", [])
        for item in input:
            if not isinstance(item, dict):
                continue

            for key, value in item.items(): # type: ignore
                if key == "content":
                    if isinstance(value, list):
                        modified = self.post_process_request_prompt(value, image_type="input_image", url_subkey=False) | modified # type: ignore

        return modified

    @override
    def remove_responses_inline_data(self, responses: 'list[dict[str, Any]]') -> bool:# noqa: ARG002
        modified = False
        def remove_from_output_array(outputs: 'list[dict[str, Any]]') -> bool:
            modified = False
            for output in outputs:
                response_type = output.get("type", "")
                if response_type == "image_generation_call" and 'result' in output:
                    output["result"] = _PayiInstrumentor._not_instrumented
                    modified = True
            return modified

        for response in responses:
            output = response.get("output", [])
            if output:
                modified |= remove_from_output_array(output)

            response_type = response.get("type", "")
            if response_type == "response.image_generation_call.partial_image" and 'partial_image_b64' in response:
                response["partial_image_b64"] = _PayiInstrumentor._not_instrumented
                modified = True
            elif response_type == "response.output_item.done":
                item = response.get("item", {})
                if item and item.get("type", "") == "image_generation_call" and 'result' in item:
                    item["result"] = _PayiInstrumentor._not_instrumented
                    modified = True
            elif response_type == "response.completed":
                completed_output = response.get("response", {}).get("output", [])
                if completed_output:
                    modified |= remove_from_output_array(completed_output)

        return modified

    @override
    def process_synchronous_response(
        self,
        response: Any,
        kwargs: Any) -> Any:

        response_dict = _model_to_dict(response)
        output = response_dict.get("output", [])
        if output:
            for o in output:
                type = o.get("type", "")
                if type != "function_call":
                    continue

                name = o.get("name", None)
                arguments = o.get("arguments", None)

                if name:
                    self.add_synchronous_function_call(name=name, arguments=arguments)

        return self.process_synchronous_response_worker(response_dict)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from ._types import (
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
    not_given,
)
from ._utils import is_given, get_async_library
from ._compat import cached_property
from ._version import __version__
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from ._exceptions import PayiError, APIStatusError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
)

if TYPE_CHECKING:
    from .resources import ingest, limits, requests, use_cases, categories
    from .resources.ingest import IngestResource, AsyncIngestResource
    from .resources.limits.limits import LimitsResource, AsyncLimitsResource
    from .resources.requests.requests import RequestsResource, AsyncRequestsResource
    from .resources.use_cases.use_cases import UseCasesResource, AsyncUseCasesResource
    from .resources.categories.categories import CategoriesResource, AsyncCategoriesResource

__all__ = ["Timeout", "Transport", "ProxiesTypes", "RequestOptions", "Payi", "AsyncPayi", "Client", "AsyncClient"]


class Payi(SyncAPIClient):
    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous Payi client instance.

        This automatically infers the `api_key` argument from the `PAYI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("PAYI_API_KEY")
        if api_key is None:
            raise PayiError(
                "The api_key client option must be set either by passing api_key to the client or by setting the PAYI_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("PAYI_BASE_URL")
        if base_url is None:
            base_url = f"https://api.pay-i.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

    @cached_property
    def limits(self) -> LimitsResource:
        """Limits"""
        from .resources.limits import LimitsResource

        return LimitsResource(self)

    @cached_property
    def ingest(self) -> IngestResource:
        """Ingest Requests"""
        from .resources.ingest import IngestResource

        return IngestResource(self)

    @cached_property
    def categories(self) -> CategoriesResource:
        """System and Custom Categories"""
        from .resources.categories import CategoriesResource

        return CategoriesResource(self)

    @cached_property
    def use_cases(self) -> UseCasesResource:
        """Use Cases"""
        from .resources.use_cases import UseCasesResource

        return UseCasesResource(self)

    @cached_property
    def requests(self) -> RequestsResource:
        from .resources.requests import RequestsResource

        return RequestsResource(self)

    @cached_property
    def with_raw_response(self) -> PayiWithRawResponse:
        return PayiWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> PayiWithStreamedResponse:
        return PayiWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"xProxy-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncPayi(AsyncAPIClient):
    # client options
    api_key: str

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async AsyncPayi client instance.

        This automatically infers the `api_key` argument from the `PAYI_API_KEY` environment variable if it is not provided.
        """
        if api_key is None:
            api_key = os.environ.get("PAYI_API_KEY")
        if api_key is None:
            raise PayiError(
                "The api_key client option must be set either by passing api_key to the client or by setting the PAYI_API_KEY environment variable"
            )
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("PAYI_BASE_URL")
        if base_url is None:
            base_url = f"https://api.pay-i.com"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

    @cached_property
    def limits(self) -> AsyncLimitsResource:
        """Limits"""
        from .resources.limits import AsyncLimitsResource

        return AsyncLimitsResource(self)

    @cached_property
    def ingest(self) -> AsyncIngestResource:
        """Ingest Requests"""
        from .resources.ingest import AsyncIngestResource

        return AsyncIngestResource(self)

    @cached_property
    def categories(self) -> AsyncCategoriesResource:
        """System and Custom Categories"""
        from .resources.categories import AsyncCategoriesResource

        return AsyncCategoriesResource(self)

    @cached_property
    def use_cases(self) -> AsyncUseCasesResource:
        """Use Cases"""
        from .resources.use_cases import AsyncUseCasesResource

        return AsyncUseCasesResource(self)

    @cached_property
    def requests(self) -> AsyncRequestsResource:
        from .resources.requests import AsyncRequestsResource

        return AsyncRequestsResource(self)

    @cached_property
    def with_raw_response(self) -> AsyncPayiWithRawResponse:
        return AsyncPayiWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncPayiWithStreamedResponse:
        return AsyncPayiWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"xProxy-api-key": api_key}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class PayiWithRawResponse:
    _client: Payi

    def __init__(self, client: Payi) -> None:
        self._client = client

    @cached_property
    def limits(self) -> limits.LimitsResourceWithRawResponse:
        """Limits"""
        from .resources.limits import LimitsResourceWithRawResponse

        return LimitsResourceWithRawResponse(self._client.limits)

    @cached_property
    def ingest(self) -> ingest.IngestResourceWithRawResponse:
        """Ingest Requests"""
        from .resources.ingest import IngestResourceWithRawResponse

        return IngestResourceWithRawResponse(self._client.ingest)

    @cached_property
    def categories(self) -> categories.CategoriesResourceWithRawResponse:
        """System and Custom Categories"""
        from .resources.categories import CategoriesResourceWithRawResponse

        return CategoriesResourceWithRawResponse(self._client.categories)

    @cached_property
    def use_cases(self) -> use_cases.UseCasesResourceWithRawResponse:
        """Use Cases"""
        from .resources.use_cases import UseCasesResourceWithRawResponse

        return UseCasesResourceWithRawResponse(self._client.use_cases)

    @cached_property
    def requests(self) -> requests.RequestsResourceWithRawResponse:
        from .resources.requests import RequestsResourceWithRawResponse

        return RequestsResourceWithRawResponse(self._client.requests)


class AsyncPayiWithRawResponse:
    _client: AsyncPayi

    def __init__(self, client: AsyncPayi) -> None:
        self._client = client

    @cached_property
    def limits(self) -> limits.AsyncLimitsResourceWithRawResponse:
        """Limits"""
        from .resources.limits import AsyncLimitsResourceWithRawResponse

        return AsyncLimitsResourceWithRawResponse(self._client.limits)

    @cached_property
    def ingest(self) -> ingest.AsyncIngestResourceWithRawResponse:
        """Ingest Requests"""
        from .resources.ingest import AsyncIngestResourceWithRawResponse

        return AsyncIngestResourceWithRawResponse(self._client.ingest)

    @cached_property
    def categories(self) -> categories.AsyncCategoriesResourceWithRawResponse:
        """System and Custom Categories"""
        from .resources.categories import AsyncCategoriesResourceWithRawResponse

        return AsyncCategoriesResourceWithRawResponse(self._client.categories)

    @cached_property
    def use_cases(self) -> use_cases.AsyncUseCasesResourceWithRawResponse:
        """Use Cases"""
        from .resources.use_cases import AsyncUseCasesResourceWithRawResponse

        return AsyncUseCasesResourceWithRawResponse(self._client.use_cases)

    @cached_property
    def requests(self) -> requests.AsyncRequestsResourceWithRawResponse:
        from .resources.requests import AsyncRequestsResourceWithRawResponse

        return AsyncRequestsResourceWithRawResponse(self._client.requests)


class PayiWithStreamedResponse:
    _client: Payi

    def __init__(self, client: Payi) -> None:
        self._client = client

    @cached_property
    def limits(self) -> limits.LimitsResourceWithStreamingResponse:
        """Limits"""
        from .resources.limits import LimitsResourceWithStreamingResponse

        return LimitsResourceWithStreamingResponse(self._client.limits)

    @cached_property
    def ingest(self) -> ingest.IngestResourceWithStreamingResponse:
        """Ingest Requests"""
        from .resources.ingest import IngestResourceWithStreamingResponse

        return IngestResourceWithStreamingResponse(self._client.ingest)

    @cached_property
    def categories(self) -> categories.CategoriesResourceWithStreamingResponse:
        """System and Custom Categories"""
        from .resources.categories import CategoriesResourceWithStreamingResponse

        return CategoriesResourceWithStreamingResponse(self._client.categories)

    @cached_property
    def use_cases(self) -> use_cases.UseCasesResourceWithStreamingResponse:
        """Use Cases"""
        from .resources.use_cases import UseCasesResourceWithStreamingResponse

        return UseCasesResourceWithStreamingResponse(self._client.use_cases)

    @cached_property
    def requests(self) -> requests.RequestsResourceWithStreamingResponse:
        from .resources.requests import RequestsResourceWithStreamingResponse

        return RequestsResourceWithStreamingResponse(self._client.requests)


class AsyncPayiWithStreamedResponse:
    _client: AsyncPayi

    def __init__(self, client: AsyncPayi) -> None:
        self._client = client

    @cached_property
    def limits(self) -> limits.AsyncLimitsResourceWithStreamingResponse:
        """Limits"""
        from .resources.limits import AsyncLimitsResourceWithStreamingResponse

        return AsyncLimitsResourceWithStreamingResponse(self._client.limits)

    @cached_property
    def ingest(self) -> ingest.AsyncIngestResourceWithStreamingResponse:
        """Ingest Requests"""
        from .resources.ingest import AsyncIngestResourceWithStreamingResponse

        return AsyncIngestResourceWithStreamingResponse(self._client.ingest)

    @cached_property
    def categories(self) -> categories.AsyncCategoriesResourceWithStreamingResponse:
        """System and Custom Categories"""
        from .resources.categories import AsyncCategoriesResourceWithStreamingResponse

        return AsyncCategoriesResourceWithStreamingResponse(self._client.categories)

    @cached_property
    def use_cases(self) -> use_cases.AsyncUseCasesResourceWithStreamingResponse:
        """Use Cases"""
        from .resources.use_cases import AsyncUseCasesResourceWithStreamingResponse

        return AsyncUseCasesResourceWithStreamingResponse(self._client.use_cases)

    @cached_property
    def requests(self) -> requests.AsyncRequestsResourceWithStreamingResponse:
        from .resources.requests import AsyncRequestsResourceWithStreamingResponse

        return AsyncRequestsResourceWithStreamingResponse(self._client.requests)


Client = Payi

AsyncClient = AsyncPayi
